#include "sv_server_filter_VolumeFilter.h"
#include "volume.h"

JNIEXPORT jint JNICALL Java_sv_server_filter_VolumeFilter_precompute
  (JNIEnv *env, jclass clazz, jstring prefix, jfloatArray rarr, 
   jfloat min, jfloat max, jint vp_xlen, jint vp_ylen, jint vp_zlen) {

#define DENSITY_RAMP_POINTS 3			/* classification ramps */
#define GRADIENT_RAMP_POINTS 4

  int DensityRampX[] =    {  0,  24, 255};
  int GradientRampX[] =   {  0,   5,  20, 221};
  float DensityRampY[] =  {0.0, 1.0, 1.0};
  float GradientRampY[] = {0.0, 0.0, 1.0, 1.0};

  const char *name = (*env)->GetStringUTFChars(env, prefix, 0);

  vpContext *vpc;	/* rendering context */

  int i;
  float factor;
  int output_fd;	/* file descriptor for classified volume (output) */
  jfloat *fdensity;	/* buffer for density data */
  unsigned char *density;
  unsigned density_size;/* size of density data */
  float density_ramp[DENSITY_MAX+1];	/* opacity as a function of density */
  float gradient_ramp[GRADIENT_MAX+1];/* opacity as a function 
					   of gradient magnitude */
  void *malloc();
  char filename[512];


  /* create a context */
  vpc = vpCreateContext();

  /* load input raw data,  describe the layout of the volume */
  vpSetVolumeSize(vpc, vp_xlen, vp_ylen, vp_zlen);
  vpSetVoxelSize(vpc, BYTES_PER_VOXEL, VOXEL_FIELDS,
		       SHADE_FIELDS, CLSFY_FIELDS);
  vpSetVoxelField(vpc, NORMAL_FIELD, NORMAL_SIZE, NORMAL_OFFSET,
			NORMAL_MAX);
  vpSetVoxelField(vpc, DENSITY_FIELD, DENSITY_SIZE, DENSITY_OFFSET,
			DENSITY_MAX);
  vpSetVoxelField(vpc, GRADIENT_FIELD, GRADIENT_SIZE, GRADIENT_OFFSET,
			GRADIENT_MAX);

  /* allocate space for the raw data */
  density_size = vp_xlen * vp_ylen * vp_zlen;
  density = malloc(density_size);
  if (density == NULL) {
     fprintf(stderr, "out of memory\n");
     exit(1);
  }

  fdensity = (*env)->GetFloatArrayElements(env, rarr, 0);

  if (min < 0 || max >= 255) {
    factor = 255/(max-min);
    for (i=0; i<density_size; i++)
      density[i] = (unsigned char)(factor * (fdensity[i]-min)); 
  } else {
    for (i=0; i<density_size; i++)
      density[i] = (unsigned char)fdensity[i];    
  }  

  /* set the classification function */
  vpRamp(density_ramp, sizeof(float), DENSITY_RAMP_POINTS, DensityRampX,
	   DensityRampY);
  vpSetClassifierTable(vpc, DENSITY_PARAM, DENSITY_FIELD, density_ramp,
			 sizeof(density_ramp));
  vpRamp(gradient_ramp, sizeof(float), GRADIENT_RAMP_POINTS, GradientRampX,
	   GradientRampY);
  vpSetClassifierTable(vpc, GRADIENT_PARAM, GRADIENT_FIELD,
			 gradient_ramp, sizeof(gradient_ramp));
  vpSetd(vpc, VP_MIN_VOXEL_OPACITY, 0.05);
  

  /* classify */
  if (vpClassifyScalars(vpc, density, density_size, DENSITY_FIELD,
			      GRADIENT_FIELD, NORMAL_FIELD) != VP_OK) {
     fprintf(stderr, "VolPack error: %s\n",
                vpGetErrorString(vpGetError(vpc)));
     return -1;
  }

  /* store the classified volume */
  sprintf(filename, "%s.cv", name);
  if ((output_fd = creat(filename, 0644)) < 0) {
     fprintf(stderr, "could not open %s\n", filename);
     return -1;
  }
  if (vpStoreClassifiedVolume(vpc, output_fd) != VP_OK) {
     fprintf(stderr, "VolPack error: %s\n",
		vpGetErrorString(vpGetError(vpc)));
     return -1;
  }
  close(output_fd);

 
  free(density);
  (*env)->ReleaseStringUTFChars(env, prefix, name); 
  (*env)->ReleaseFloatArrayElements(env, rarr, fdensity, 0);
  return 0;
}

